@@ -12,12 +12,17 @@ Example
 ...     ''')
 ...     print(f.read('F'))
 4*p1.p2*p3.p4-4*p1.p3*p2.p4+4*p1.p4*p2.p3
+
 """
 
-from .formlink import FormLink
+if False:
+    from typing import Optional, Sequence, Union  # noqa: F401
+
+from .formlink import FormError, FormLink  # noqa: F401
 
 
 def open(args=None, keep_log=False):
+    # type: (Optional[Union[str, Sequence[str]]], Union[bool, int]) -> FormLink
     """Open a connection to FORM and return a link object.
 
     Open a connection to a new FORM process and return a
@@ -30,19 +35,20 @@ def open(args=None, keep_log=False):
     >>> with form.open() as formlink:
     ...     pass  # use formlink ...
 
-    The optional argument `args` is for the FORM command, a string or
+    The optional argument ``args`` is for the FORM command, a string or
     a sequence of strings. For example '/path/to/form' or ['tform', '-w4'].
-    By default, the value of the environment variable `$FORM` is used if set,
+    By default, the value of the environment variable ``$FORM`` is used if set,
     otherwise 'form' will be used.
 
-    The other argument `keep_log` indicates whether the log from FORM is kept
+    The other argument ``keep_log`` indicates whether the log from FORM is kept
     and used as detailed information when an error occurs. If the value
     is >= 2, it specifies the maximum number of lines for the scrollback.
     The default value is False.
 
     Note
     ----
-    In the current implementation, keep_log=True may cause a dead lock when the
-    listing of the input is enabled and very long input is sent to FORM.
+    In the current implementation, ``keep_log=True`` may cause a dead lock when
+    the listing of the input is enabled and very long input is sent to FORM.
+
     """
     return FormLink(args, keep_log)
