@@ -1,6 +1,7 @@
+"""Routines for connections to FORM."""
+
 import collections
 import errno
-import fcntl
 import os
 import select
 import shlex
@@ -8,63 +9,18 @@ import subprocess
 import sys
 
 from .datapath import get_data_path
+from .ioutil import PushbackReader, set_nonblock
+from .six import string_types
 
-_PY3 = sys.version_info[0] >= 3
-_PY32 = sys.version_info >= (3, 2, 0)
+if False:
+    from typing import Any, IO, MutableSequence, Optional, Sequence, Tuple, Union, overload  # noqa: E501, F401
+if True:
+    def overload(f):  # type: ignore  # noqa: D103, F811
+        return None
 
-if not _PY3:
-    # Python 2
-    def is_string(obj):
-        """Return True if the given object is a string."""
-        return isinstance(obj, basestring)
-else:
-    # Python 3
-    def is_string(obj):
-        """Return True if the given object is a string."""
-        return isinstance(obj, str)
-
-def set_nonblock(fd):
-    """Set the given file descriptor to non-blocking mode."""
-    fcntl.fcntl(fd,
-                fcntl.F_SETFL,
-                fcntl.fcntl(fd, fcntl.F_GETFL) | os.O_NONBLOCK)
-
-class PushbackReader(object):
-    """A wrapper class for streams that allows data to be pushed back into the
-       stream."""
-
-    def __init__(self, raw):
-        self._raw = raw
-        self._buf = ''
-
-    def close(self):
-        """Close the stream."""
-        self._raw.close()
-
-    def fileno(self):
-        """Return the file descriptor."""
-        return self._raw.fileno()
-
-    def read(self):
-        """Read data from the stream."""
-        s = self._buf + self._raw.read()
-        self._buf = ''
-        return s
-
-    def unread(self, s):
-        """Push back the given string to the internal buffer, which will be used
-           for the next `read()` or `read0()`."""
-        self._buf = s + self._buf
-
-    def read0(self):
-        """Read the pushed-back data. No call to the underlying raw stream's
-           `read()`."""
-        s = self._buf
-        self._buf = ''
-        return s
 
 class FormLink(object):
-    """An object representing a connection to FORM."""
+    """Connection to a FORM process."""
 
     # The input file for FORM.
     _INIT_FRM = get_data_path('form', 'init.frm')
@@ -75,55 +31,87 @@ class FormLink(object):
     _PROMPT = '\n__READY__\n'
 
     def __init__(self, args=None, keep_log=False):
-        """Initialize a connection to a FORM process."""
+        # type: (Optional[Union[str, Sequence[str]]], Union[bool, int]) -> None
+        """Open a connection to a FORM process."""
         self._closed = True
-        self._head = None
-        self._log = None
-        self._childpid = None
-        self._formpid = None
-        self._parentin = None
-        self._parentout = None
-        self._loggingin = None
+        self._head = None       # type: Optional[str]
+        self._log = None        # type: Optional[MutableSequence[str]]
+        self._childpid = None   # type: Optional[int]
+        self._formpid = None    # type: Optional[int]
+        self._parentin = None   # type: Optional[PushbackReader]
+        self._parentout = None  # type: Optional[IO[str]]
+        self._loggingin = None  # type: Optional[PushbackReader]
         self.open(args, keep_log)
 
+    def __del__(self):
+        # type: () -> None
+        """Destructor.
+
+        Free the connection to the FORM process if it still exists. Since in
+        general when the destructor is called in the garbage collection is hard
+        to be determined (and may not be called at all until the program
+        finishes), it is advisable to use the "with" statement.
+        """
+        # Ignore possible errors caused by subtle timing problems.
+        try:
+            self.close()
+        except Exception:
+            pass
+
     def __enter__(self):
+        # type: () -> FormLink
+        """Enter the runtime context."""
         return self
 
     def __exit__(self, exc_type, exc_value, traceback):
+        # type: (Any, Any, Any) -> None
+        """Exit the runtime context."""
         self.close()
 
     def open(self, args=None, keep_log=False):
+        # type: (Optional[Union[str, Sequence[str]]], Union[bool, int]) -> None
         """Open a connection to FORM.
 
-        Open a connection to a new FORM process. The opened connection should be
-        closed by `close()`. Since `open()` is called from the initializer (or
-        `form.open()`), this can be guaranteed by use of the "with" statement:
+        Open a connection to a new FORM process. The opened connection should
+        be closed by :meth:`close`, which can be guaranteed by use of the
+        "with" statement:
 
         >>> import form
         >>> with form.open() as formlink:
         ...     pass  # use formlink ...
 
-        The optional argument `args` is for the FORM command, a string or
+        If this method is called for a link object that has an established
+        connection to a FORM process, then the existing connection will be
+        closed and a new connection will be created.
+
+        The optional argument ``args`` is for the FORM command, a string or
         a sequence of strings. For example '/path/to/form' or ['tform', '-w4'].
-        The default value is 'form'.
+        By default, the value of the environment variable ``$FORM`` is used if
+        set, otherwise 'form' will be used.
 
-        The other argument `keep_log` indicates whether the log from FORM is
-        kept and used as detailed information when an error occurs. If the value
-        is >= 2, it specifies the maximum number of lines for the scrollback.
-        The default value is False.
+        The other argument ``keep_log`` indicates whether the log from FORM is
+        kept and used as detailed information when an error occurs.
+        If the value is >= 2, it specifies the maximum number of lines for
+        the scrollback. The default value is False.
+
+        Note
+        ----
+        In the current implementation, ``keep_log=True`` may cause a dead lock
+        when the listing of the input is enabled and very long input is sent to
+        FORM.
 
-        Caveats
-        -------
-        In the current implementation, keep_log=True may cause a dead lock when
-        the listing of the input is enabled and very long input is sent to FORM.
         """
         if args is None:
-            args = 'form'
-
-        if is_string(args):
+            if 'FORM' in os.environ:
+                args = os.environ['FORM']
+            else:
+                args = 'form'
+        if isinstance(args, string_types):
             args = shlex.split(args)  # Split the arguments.
         elif isinstance(args, (list, tuple)):
             args = list(args)  # As a modifiable mutable object.
+        else:
+            raise TypeError("invalid args = {0}".format(args))
 
         self.close()
 
@@ -171,14 +159,14 @@ class FormLink(object):
 
             set_nonblock(fd_parentin)
             set_nonblock(fd_loggingin)
-            parentin = PushbackReader(parentin)
-            loggingin = PushbackReader(loggingin)
 
             self._closed = False
             self._head = head
             if keep_log:
                 if keep_log >= 2:
-                    self._log = collections.deque(maxlen=keep_log)
+                    log = collections.deque(maxlen=keep_log)  # type: Any
+                    self._log = log  # hack typeshed for Python 2
+                    assert self._log is not None
                     self._log.append(head)
                 else:
                     self._log = []
@@ -189,11 +177,16 @@ class FormLink(object):
                 parentout.write('#-\n')
             self._childpid = pid
             self._formpid = formpid
-            self._parentin = parentin
+            self._parentin = PushbackReader(parentin)
             self._parentout = parentout
-            self._loggingin = loggingin
+            self._loggingin = PushbackReader(loggingin)
         else:
             # child process
+
+            # NOTE: Coverage fails to collect data from child processes when
+            #       the os.fork+os._exit pattern is used.
+            #       https://bitbucket.org/ned/coveragepy/issues/310
+
             os.close(fd_parentout)
             os.close(fd_parentin)
             os.close(fd_loggingin)
@@ -204,7 +197,11 @@ class FormLink(object):
             args.append('{0},{1}'.format(fd_childin, fd_childout))
             args.append(FormLink._INIT_FRM)
 
-            if not _PY32:
+            # In Python 3.2, subprocess.Popen() on UNIX changed the default
+            # value for close_fds from False to True, in order to stop leaking
+            # file descriptors. File descriptors to be kept open should be
+            # specified by pass_fds.
+            if sys.version_info[0:2] < (3, 2):
                 subprocess.call(args, shell=False)
             else:
                 subprocess.call(args, shell=False,
@@ -218,20 +215,24 @@ class FormLink(object):
             os._exit(0)
 
     def close(self):
+        # type: () -> None
         """Close the connection to FORM.
 
-        Close the connection to the FORM process established by `open()`. The
-        user should call this method after use of each FormLink object.
+        Close the connection to the FORM process established by :meth:`open`.
+        Do nothing if the connection is already closed. The user should call
+        this method after use of a link object, which is usually guaranteed by
+        use of the "with" statement.
         """
         self._close()
 
-    def kill(self):
-        """Kill the FORM process and close the connection."""
-        self._close(kill=-1)  # Kill it immediately.
-#       self._close(term=-1, kill=1)
-
     def _close(self, term=False, kill=False):
+        # type: (float, float) -> None
         if not self._closed:
+            assert self._childpid is not None
+            assert self._formpid is not None
+            assert self._parentin is not None
+            assert self._parentout is not None
+            assert self._loggingin is not None
             try:
                 # We ignore broken pipes.
                 try:
@@ -245,23 +246,25 @@ class FormLink(object):
                     import signal
                     import time
 
-                    # When a non-zero `term` or `kill` is given, we first wait
-                    # for the child to finish within the duration. If not, stop
-                    # it by SIGTERM/SIGKILL. If both `term` and `kill` are
-                    # non-zero, we first try SIGTERM, and then SIGKILL.
+                    # When a non-zero ``term`` or ``kill`` is given, we first
+                    # wait for the child to finish within the duration. If not,
+                    # stop it by SIGTERM/SIGKILL. If both ``term`` and ``kill``
+                    # are non-zero, we first try SIGTERM, and then SIGKILL.
                     # To stop the FORM process, we do not use setpgrp() and
                     # killpg() for the child, but directly use kill() for the
                     # FORM process. We expect that then the child process can
                     # finish shortly.
 
                     def wait(timeout):  # timeout <= 0 means no wait
+                        # type: (float) -> bool
                         # Wait for the child to finish.
+                        assert self._childpid is not None
                         t = 0.0
                         dt = 0.01
                         if timeout > 0:
                             dt = min(timeout, dt)
                         while True:
-                            pid, err = os.waitpid(self._childpid, os.WNOHANG)
+                            pid, _ = os.waitpid(self._childpid, os.WNOHANG)
                             if pid:
                                 return False
                             if t >= timeout:
@@ -299,35 +302,137 @@ class FormLink(object):
                 self._parentout = None
                 self._loggingin = None
 
+    def kill(self):
+        # type: () -> None
+        """Kill the FORM process and close the connection."""
+        self._close(kill=-1)  # Kill it immediately.
+#       self._close(term=-1, kill=1)
+
     def write(self, script):
+        # type: (str) -> None
         """Send a script to FORM.
 
         Write the given script to the communication channel to FORM. It could
-        be buffered and so FORM may not execute the sent script until `flush()`
-        or `read()` is called.
+        be buffered and so FORM may not execute the sent script until
+        :meth:`flush` or :meth:`read` is called.
         """
         if self._closed:
             raise IOError('tried to write to closed connection')
         script = script.strip()
         if script:
+            assert self._parentout is not None
             self._parentout.write(script)
             self._parentout.write('\n')
 
     def flush(self):
+        # type: () -> None
         """Flush the channel to FORM.
 
-        Flush the communication channel to FORM.
+        Flush the communication channel to FORM. Because :meth:`write` is
+        buffered and :meth:`read` is a blocking operation, this method is used
+        for asynchronous execution of FORM scripts.
         """
         if self._closed:
             raise IOError('tried to flush closed connection')
+        assert self._parentout is not None
         self._parentout.flush()
 
-    def read(self, *names):
-        """Read results from FORM.
+    @overload
+    def read(self, name):  # noqa: D102
+        # type: (str) -> str
+        pass  # pragma: no cover
 
-        Wait for a response of FORM to obtain the results specified by the given
-        names and return a corresponding string or (nested) list of strings.
-        Objects to be read from FORM are expressions, $-variables and
+    @overload  # noqa: F811
+    def read(self, name1, name2):  # noqa: D102
+        # type: (str, str) -> Tuple[str, str]
+        pass  # pragma: no cover
+
+    @overload  # noqa: F811
+    def read(self, name1, name2, name3):  # noqa: D102
+        # type: (str, str, str) -> Tuple[str, str, str]
+        pass  # pragma: no cover
+
+    @overload  # noqa: F811
+    def read(self, name1, name2, name3, name4):  # noqa: D102
+        # type: (str, str, str, str) -> Tuple[str, str, str, str]
+        pass  # pragma: no cover
+
+    @overload  # noqa: F811
+    def read(self, name1, name2, name3, name4, name5):  # noqa: D102
+        # type: (str, str, str, str, str) -> Tuple[str, str, str, str, str]
+        pass  # pragma: no cover
+
+    @overload  # noqa: F811
+    def read(self, name1, name2, name3, name4, name5, name6):  # noqa: D102
+        # type: (str, str, str, str, str, str) -> Tuple[str, str, str, str, str, str]  # noqa: E501
+        pass  # pragma: no cover
+
+    @overload  # noqa: F811
+    def read(self, name1, name2, name3, name4, name5, name6, name7):  # noqa: D102, E501
+        # type: (str, str, str, str, str, str, str) -> Tuple[str, str, str, str, str, str, str]  # noqa: E501
+        pass  # pragma: no cover
+
+    @overload  # noqa: F811
+    def read(self, name1, name2, name3, name4, name5, name6, name7, name8):  # noqa: D102, E501
+        # type: (str, str, str, str, str, str, str, str) -> Tuple[str, str, str, str, str, str, str, str]  # noqa: E501
+        pass  # pragma: no cover
+
+    @overload  # noqa: F811
+    def read(self, name1, name2, name3, name4, name5, name6, name7, name8, name9):  # noqa: D102, E501
+        # type: (str, str, str, str, str, str, str, str, str) -> Tuple[str, str, str, str, str, str, str, str, str]  # noqa: E501
+        pass  # pragma: no cover
+
+    @overload  # noqa: F811
+    def read(self, names):  # noqa: D102
+        # type: (Tuple[str]) -> str
+        pass  # pragma: no cover
+
+    @overload  # noqa: F811
+    def read(self, names):  # noqa: D102
+        # type: (Tuple[str, str]) -> Tuple[str, str]
+        pass  # pragma: no cover
+
+    @overload  # noqa: F811
+    def read(self, names):  # noqa: D102
+        # type: (Tuple[str, str, str]) -> Tuple[str, str, str]
+        pass  # pragma: no cover
+
+    @overload  # noqa: F811
+    def read(self, names):  # noqa: D102
+        # type: (Tuple[str, str, str, str]) -> Tuple[str, str, str, str]
+        pass  # pragma: no cover
+
+    @overload  # noqa: F811
+    def read(self, names):  # noqa: D102
+        # type: (Tuple[str, str, str, str, str]) -> Tuple[str, str, str, str, str]  # noqa: E501
+        pass  # pragma: no cover
+
+    @overload  # noqa: F811
+    def read(self, names):  # noqa: D102
+        # type: (Tuple[str, str, str, str, str, str]) -> Tuple[str, str, str, str, str, str]  # noqa: E501
+        pass  # pragma: no cover
+
+    @overload  # noqa: F811
+    def read(self, names):  # noqa: D102
+        # type: (Tuple[str, str, str, str, str, str, str]) -> Tuple[str, str, str, str, str, str, str]  # noqa: E501
+        pass  # pragma: no cover
+
+    @overload  # noqa: F811
+    def read(self, *names):  # noqa: D102
+        # type: (str) -> Sequence[str]
+        pass  # pragma: no cover
+
+    @overload  # noqa: F811
+    def read(self, names):  # noqa: D102
+        # type: (Sequence[str]) -> Sequence[str]
+        pass  # pragma: no cover
+
+    def read(self, *names):  # type: ignore  # noqa: F811
+        r"""Read results from FORM.
+
+        Wait for a response of FORM to obtain the results specified by
+        the given names and return a corresponding string or (nested) list of
+        strings. Objects to be read from FORM are expressions, $-variables and
         preprocessor variables.
 
         ========== =============================
@@ -336,7 +441,7 @@ class FormLink(object):
           "F"        expression F
           "$x"       $-variable $x
           "$x[]"     factorized $-variable $x
-          "`A'"      preprocessor variable A
+          "\`A'"     preprocessor variable A
         ========== =============================
 
         Note that the communication for the reading is performed within the
@@ -344,7 +449,7 @@ class FormLink(object):
         ".sort" to get the correct result.
 
         If non-string objects are passed, they are considered as sequences, and
-        the return value becomes the list corresponding to the arguments. If
+        the return value becomes a list corresponding to the arguments. If
         a sequence is passed as the argument to this method, it is guaranteed
         that the return value is always a list:
 
@@ -371,29 +476,41 @@ class FormLink(object):
         ['a1', ['a2', 'a3']]
 
         >>> f.close()
+
+        Note
+        ----
+        Currently nested inputs are not supported for static typing. You need
+        ``# type: ignore`` or other hacks.
+
         """
         if self._closed:
             raise IOError('tried to read from closed connection')
 
-        if len(names) == 1 and not is_string(names[0]):
+        if len(names) == 1 and not isinstance(names[0], string_types):
             names = tuple(names[0])
             if len(names) == 1:
                 return [self.read(*names)]  # Guarantee to return a list.
             else:
                 return self.read(*names)
 
-        if any(not is_string(x) for x in names):
+        if any(not isinstance(x, string_types) for x in names):
             return [self.read(x) for x in names]
 
+        assert self._parentin is not None
+        assert self._parentout is not None
+        assert self._loggingin is not None
+
         for e in names:
             if len(e) >= 2 and e[0] == '`' and e[-1] == "'":
-                self._parentout.write('#toexternal "{0}{1}"\n'.format(e, self._END_MARK))
+                self._parentout.write(
+                    '#toexternal "{0}{1}"\n'.format(e, self._END_MARK))
             elif len(e) >= 3 and e[0] == '$' and e[-2:] == '[]':
                 # Special syntax "$x[]" for factorized $-variables.
-                # NOTE: (1) isfactorized($x) is zero when $x is 0 or $x has only
-                #           one factor even after FactArg is performed.
+                # NOTE: (1) isfactorized($x) is zero when $x is 0 or $x has
+                #           only one factor even after FactArg is performed.
                 #       (2) `$x[0]' is accessible even if FactArg has not been
-                #           performed. Use `$x[0]' rather than isfactorized($x).
+                #           performed. Use `$x[0]' rather than
+                #           `isfactorized($x)`.
                 #       (3) `$x[1]' is not accessible (segfault) with versions
                 #           before Sep  3 2015, if $x has only one factor and
                 #           `$x[0]' gives 1.
@@ -411,24 +528,29 @@ class FormLink(object):
                     "#endif\n"
                     "#endif\n"
                     "#toexternal \"{1}\"\n"
-                    ).format(e[1:-2], self._END_MARK))
+                ).format(e[1:-2], self._END_MARK))
             elif len(e) >= 1 and e[0] == '$':
-                self._parentout.write('#toexternal "%${1}",{0}\n'.format(e, self._END_MARK))
+                self._parentout.write(
+                    '#toexternal "%${1}",{0}\n'.format(e, self._END_MARK))
             else:
-                self._parentout.write('#toexternal "%E{1}",{0}\n'.format(e, self._END_MARK))
+                self._parentout.write(
+                    '#toexternal "%E{1}",{0}\n'.format(e, self._END_MARK))
         self._parentout.write('#redefine FORMLINKLOOPVAR "0"')
         self._parentout.write(self._PROMPT)
         self._parentout.flush()
 
         result = []
         out = self._parentin.read0()
-        for e in names:
+        out_start = 0  # start position for searching _END_MARK.
+        for _e in names:
             while True:
-                i = out.find(self._END_MARK)
+                i = out.find(self._END_MARK, out_start)
                 if i >= 0:
                     result.append(out[:i])
-                    out = out[i+self._END_MARK_LEN:]
+                    out = out[i + self._END_MARK_LEN:]
+                    out_start = 0
                     break
+                out_start = max(len(out) - self._END_MARK_LEN, 0)
 
                 r, _, _ = select.select((self._parentin, self._loggingin),
                                         (), ())
@@ -441,13 +563,14 @@ class FormLink(object):
                             if self._log is not None:
                                 self._log.extend(msgs)
                             for msg in msgs:
-                                if msg.find('-->') >= 0 or msg.find('==>') >= 0:
+                                if (msg.find('-->') >= 0 or
+                                        msg.find('==>') >= 0):
                                     if self._log:
                                         msg += '\n'
                                         msg += '\n'.join(self._log)
                                     self.close()
-                                    raise RuntimeError(msg)
-                        self._loggingin.unread(s[i+1:])
+                                    raise FormError(msg)
+                        self._loggingin.unread(s[i + 1:])
                 if self._parentin in r:
                     out += (self._parentin.read()
                             .replace('\n', '')
@@ -465,23 +588,27 @@ class FormLink(object):
 
     @property
     def closed(self):
+        # type: () -> bool
         """Return True if the connection is closed."""
         return self._closed
 
     @property
     def head(self):
+        # type: () -> str
         """Return the first line of the FORM output."""
+        assert self._head is not None
         return self._head
 
     @property
     def _dateversion(self):
+        # type: () -> int
+        """Return the build/revision date as an integer "yyyymmdd"."""
         import re
         if self._head:
-            m = re.search(r'(?<=\()(.*)(?=\))', self._head)
-            if m:
-                dv = m.group(0)
-                s = dv.split()
-                if len(s) == 3:
+            ma = re.search(r'(?<=\()(.*)(?=\))', self._head)
+            if ma:
+                s = re.split(r'[, ]+', ma.group(0))
+                if len(s) >= 3:
                     # month
                     month_names = ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
                                    'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec')
@@ -499,3 +626,11 @@ class FormLink(object):
                                         return y * 10000 + m * 100 + d
             raise ValueError('failed to parse "{0}"'.format(self._head))
         raise ValueError('no first line')
+
+
+class FormError(RuntimeError):
+    """FORM stopped by an error.
+
+    This exception is raised when :meth:`read() <form.FormLink.read>` finds
+    the FORM process stopped by some error.
+    """
